"""
    Traio: A simple asyncio wrapper attempting to look like Trio

    When going deeper and deeper with asyncio, and managing a lot of tasks
    in parallel, you notice that on top of having a lot to deal with
    to keep an eye on all your task, but you also end up always doing the
    same kind of boiler plate...

    Traio (as Trio on asyncio) let you use asyncio with the philosophy of Trio.
    This is *not* a replacement for Trio: if you do a full Trio-like project,
    just switch to it!
    Trio is just awesome, but in some cases you get stuck with asyncio,
    but still want to have a code you can read and manage...
"""

from .nursery import Nursery
from .task import TaskException

__version__ = "0.1.1"

__all__ = [
    'Nursery',
    'TaskException'
]



"""
Test various misuse of this library.
"""

import pytest

from traio import Nursery


@pytest.mark.parametrize('arg', [
    None,
    1,
    'hello',
    lambda x: x + 1
])
@pytest.mark.asyncio
async def test_task_not_awaitable(arg):
    """Call start_soon with different non-awaitable objects"""
    with pytest.raises(OSError):
        async with Nursery() as n:
            n.start_soon(arg)


@pytest.mark.asyncio
async def test_sync_ctx_manager():
    """Calling nursery as a synchronous context manager"""
    with pytest.raises(RuntimeError):
        with Nursery():
            pass



"""
AsyncTask wrapping an asyncio task, with a bit of boiler plate.
"""

import asyncio
import time
from typing import Awaitable


class TaskException(Exception):
    """Exception raised inside a Task will be wrapped"""

    def __init__(self, cause):
        super().__init__()
        self.__cause__ = cause


class AsyncTask:
    """
    Task convenient object used by nursery
    """

    def __init__(self, nursery, awaitable: Awaitable, *,
                 cancel_timeout=1, bubble=True, master=False, name=None):
        """
        Create a Task to be executed
        :param nursery: Parent nursery
        :param awaitable: awaitable we are wrapping
        :param cancel_timeout: When cancelling, ensure the coroutine has finished
        :param bubble: an exception in this task will pop in the nursery
        :param master: when this task finishes, cancel nursery
        :param name: convenient name for logging
        """
        self.nursery = nursery
        self.cancel_timeout = cancel_timeout
        self.bubble = bubble
        self.master = master
        self.name = name

        self.start_time = time.time()
        self.future = asyncio.ensure_future(awaitable)
        self.future.add_done_callback(self._cleanup)

    def __repr__(self):
        return self.name if self.name is not None else str(self.future)

    def _cleanup(self, fut: asyncio.Future):
        try:
            fut.result()
        except asyncio.CancelledError:
            self.nursery.logger.info(
                'task `%s` got cancelled', self)
        except Exception as ex:  # pylint: disable=broad-except
            self.nursery.logger.error(
                'task `%s` got an exception: %s', self, ex)
            if self.bubble:
                self.nursery.cancel(TaskException(ex))
        else:
            self.nursery.logger.info('task `%s` done', self)
        finally:
            if self.master:
                self.nursery.cancel()
            self.nursery.remove_task(self)

    # --- API ---

    async def join(self):
        """
        Wait for internal future to terminate
        :returns: whatever the internal things returned
        """
        return await self.future

    def done(self):
        """Internal future is done"""
        return self.future.done()

    def cancel(self):
        """Perform cancellation on internal future"""
        self.future.cancel()

    async def ensure_cancelled(self):
        """
        Ensure cancelled task has finished
        (a coro may catch cancel error for cleanup...)
        """
        self.cancel()
        try:
            await asyncio.wait_for(
                # Since python 3.7, after timeout the wait_for API
                # will try to cancel and await the future... which may block forever!
                asyncio.shield(self.future),
                self.cancel_timeout
            )
        except asyncio.CancelledError:
            pass
        except asyncio.TimeoutError as ex:
            self.nursery.logger.error(
                'task `%s` could not be cancelled in time', self)
            raise OSError(
                'Could not cancel task `{}`!!! Check your code!'.format(self)) from ex



"""
Testing the `bubble` flag feature of `Nursery.start_soon`
"""

import asyncio
import time

import pytest

from tests import run10
from traio import Nursery


@pytest.mark.asyncio
async def test_no_bubble():
    """Test iif no bubble does not cancel nursery"""
    async def trivial():
        await asyncio.sleep(0.01)
        raise Exception('not interesting')

    before = time.time()

    with pytest.raises(TimeoutError):
        async with Nursery(timeout=0.5) as n:
            n.start_soon(run10())
            n.start_soon(trivial(), bubble=False)

    after = time.time()

    assert (after - before) > 0.4


@pytest.mark.asyncio
async def test_no_bubble_master():
    """Test if no bubble does not raise, but cancels because of master=True"""
    async def trivial():
        await asyncio.sleep(0.01)
        raise Exception('not interesting')

    before = time.time()

    async with Nursery(timeout=0.5) as n:
        n.start_soon(run10())
        n.start_soon(trivial(), bubble=False, master=True)

    after = time.time()

    assert (after - before) < 0.2



"""
Nursery timeout tests
"""
import pytest
import time

from tests import run10
from traio import Nursery


@pytest.mark.asyncio
async def test_timeout():
    """Simple timeout"""
    before = time.time()

    with pytest.raises(TimeoutError):
        async with Nursery(timeout=0.1) as n:
            n.start_soon(run10())

    after = time.time()
    assert (after - before) < 0.5


@pytest.mark.asyncio
async def test_timeout_override():
    """Override timeout value"""
    before = time.time()

    with pytest.raises(TimeoutError):
        async with Nursery(timeout=5) as n:
            n.start_soon(run10())
            n.timeout = 0.1

    after = time.time()
    assert (after - before) < 0.5



"""
Various task & nursery cancellation tests
"""

import asyncio
import time

import pytest

from tests import run10
from traio import Nursery
from traio.nursery import State


@pytest.mark.asyncio
async def test_task_catches_cancel():
    """A nasty task catches all exceptions"""
    async def nasty():
        while True:
            try:
                await asyncio.sleep(1)
                print('loop')
            except asyncio.CancelledError:
                # Prevent the bloody cancel!
                pass

    before = time.time()

    with pytest.raises(OSError):
        async with Nursery(timeout=0.1) as n:
            n.start_soon(nasty(), cancel_timeout=0.5)

    after = time.time()
    assert 0.4 < (after - before) < 1


@pytest.mark.asyncio
async def test_external_cancel():
    """Raise an external cancellation"""
    async def run_me():
        n = Nursery()
        async with n:
            n.start_soon(run10())

        assert n.state == State.CANCELLED

    task = asyncio.ensure_future(run_me())

    try:
        await asyncio.wait_for(asyncio.shield(task), 0.1)
    except asyncio.TimeoutError:
        task.cancel()
        with pytest.raises(asyncio.CancelledError):
            await task
    else:
        raise Exception('should have raised')


@pytest.mark.asyncio
async def test_external_cancel_nasty():
    """Raise an external cancellation with task which fails cancelling"""
    async def nasty():
        try:
            await asyncio.sleep(10)
        except asyncio.CancelledError:
            # Something bad happens
            raise ValueError('boom')

    async def run_me():
        n = Nursery()
        async with n:
            n.start_soon(nasty())

        assert n.state == State.CANCELLED

    task = asyncio.ensure_future(run_me())

    try:
        await asyncio.wait_for(task, 0.1)
    except asyncio.TimeoutError:
        with pytest.raises(asyncio.CancelledError):
            await task
    else:
        raise Exception('should have raised')


@pytest.mark.asyncio
async def test_internal_cancel():
    """Test an internal cancellation"""
    before = time.time()

    async with Nursery() as n:
        n.start_soon(run10())
        await asyncio.sleep(0.2)
        n.cancel()

    after = time.time()

    assert (after - before) < 0.5


@pytest.mark.asyncio
async def test_cancel_not_started():
    """cancel an unstarted nursery"""
    n = Nursery()

    with pytest.raises(AssertionError):
        n.cancel()


@pytest.mark.asyncio
async def test_cancelling_going_bad():
    """Test cancelling a pending task, but things go wrong..."""
    async def nasty():
        try:
            await asyncio.sleep(10)
        except asyncio.CancelledError:
            # Something bad happens
            raise ValueError('boom')

    with pytest.raises(TimeoutError):
        async with Nursery(timeout=0.5) as n:
            n.start_soon(nasty())

    await asyncio.sleep(0.1)



"""
Implement a kind-of Nursery like what you would get of Trio.
Except:
    - we do it on top of asyncio
    - we do not pretend implementing an equivalent of Trio; just the philosophy!
    - no pretensions; if you want more Trio-like, just switch to Trio!
"""
import asyncio
import logging
from enum import Enum
from typing import Awaitable

from .task import AsyncTask


class State(Enum):
    """States of a nursery instance"""
    INIT = 0
    STARTED = 1
    DONE = 2
    CANCELLED = 3
    ERROR = 4


class Nursery:
    """
    Trio-like nursery (or at least a very light & dumb implementation...)

    At least it runs on top of asyncio, not "instead of" it.

    You can use it in 2 ways:

    1) With an environment, automatically joining the nursery at exit

    ```
    async with Nursery(...) as nursery:
        nursery.start_soon(my_task1)
        nursery.start_soon(my_task2)
    ```

    2) Without it:
    ```
    nursery = Nursery(...)
    nursery.start_soon(my_task1)
    nursery.start_soon(my_task2)
    nursery.start()

    [...]
    await nursery.join()
    ```

    At any point in time, the nursery can be cancelled (with or without an exception).
    This will cause:
        - the stopping and cleaning up of all tasks
        - the nursery to be marked as done (join() will return)
    """

    # pylint: disable=too-many-instance-attributes
    def __init__(self, logger=None, timeout=None, name=None):
        """
        Create a nursery.

        :param logger: Can pass a logger. By default using `traio` logger
        :param timeout: if timeout is specified
        :param name:
        """
        self.logger = logger or logging.getLogger('traio')

        self._pending_tasks = []

        assert timeout is None or timeout >= 0, \
            'timeout should be a positive integer or (0, None) when disabled'
        self._timeout_task = None
        self.timeout = 0
        self._timeout = timeout

        self._name = name or '-'
        self._done = asyncio.Future()
        self.state = State.INIT

    def __repr__(self):
        """For printing"""
        return self._name

    async def __aenter__(self):
        self.start()
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        if exc_type:
            # An exception occurred: cleanup
            self.cancel(exc_val)
        await self.join()

    def __enter__(self):
        """Protect against calling as regular context manager"""
        raise RuntimeError(
            "asynchronous context manager, use 'async with Nursery(...)'!"
        )

    def __exit__(self, *_):  # pragma: no cover
        assert False, 'This should never be called'

    async def _timeout_handler(self):
        await asyncio.sleep(self.timeout)
        self.cancel(TimeoutError())

    def remove_task(self, task: AsyncTask):
        """Remove a task from the nursery"""
        if task in self._pending_tasks:
            self._pending_tasks.remove(task)

        if not self._pending_tasks and not self._done.done():
            self.cancel()

    @property
    def timeout(self) -> float:
        """Get current timeout value"""
        return self._timeout

    @timeout.setter
    def timeout(self, value: float):
        """
        Reset timeout to value
        """
        self._timeout = value
        if self._timeout_task:
            self._timeout_task.cancel()
            self._timeout_task = None
        if self.timeout > 0:
            self._timeout_task = self.start_soon(
                self._timeout_handler(), name='nursery timeout')

    def cancel(self, exception: Exception = None):
        """
        Cancel nursery.
        This will stop whatever was started by the nursery,
        and raise given Exception if needed.
        :param exception: Exception to be raised
        """
        assert self.state != State.INIT, 'cannot cancel before even starting'

        for task in self._pending_tasks:
            if not task.done():
                self.logger.info(
                    'cancelling active `%s` task from nursery `%s`', task, self)
                task.cancel()

        if not self._done.done():
            if exception:
                self.logger.warning(
                    'cancelling nursery `%s` with %s: %s',
                    self, exception.__class__.__name__, exception
                )
                self._done.set_exception(exception)
            else:
                self.logger.info('cancelling nursery `%s`', self)
                self._done.set_result(None)

    def start(self):
        """
        Start the Nursery. If any timeout is configured, we start counting from now.
        """
        assert self.state == State.INIT, 'can only start a nursery once'
        self.state = State.STARTED

        if self._timeout:
            # Force timeout reset to now
            self.timeout = self._timeout

    async def join(self):
        """Await for all tasks to be finished, or an error to go through"""
        assert self.state == State.STARTED, 'can only join a started nursery'

        if not self._pending_tasks and not self._done.done():
            # There is no task registered or left.
            self._done.set_result(None)

        try:
            await self._done
            self.state = State.DONE
        except asyncio.CancelledError:
            self.logger.info('nursery `%s` cancelled from outside!', self)
            self.state = State.CANCELLED
            raise
        finally:
            if self.state == State.STARTED:
                self.state = State.ERROR

            # We may still have pending tasks if the Nursery is cancelled
            for task in self._pending_tasks:
                if not task.done():
                    try:
                        await task.ensure_cancelled()
                    except Exception as ex:  # pylint: disable=broad-except
                        # Too late for raising... and we need to move on cleaning other tasks!
                        self.logger.error(
                            'task `%s` failed to cancelled with exception: %s %s',
                            task, ex.__class__.__name__, ex)

    def start_soon(
            self, awaitable: Awaitable, *,
            bubble=True, cancel_timeout=1, master=False, name=None
    ) -> AsyncTask:
        """
        Start a task on the nursery.

        :param awaitable: Something to be awaited. Can be a future or a coroutine
        :param bubble: errors in the task will cancel nursery
        :param cancel_timeout: Time we allow for cancellation
            (if the task wants to catch it and do cleanup)
        :param master: If a master task is done, nursery is cancelled
        :param name: task name (for logging)
        :return: the task created
        """
        if asyncio.iscoroutine(awaitable) or asyncio.isfuture(awaitable):
            # This is already an awaitable object
            fut = awaitable
        else:
            raise OSError('this thing is not awaitable: {}!'.format(awaitable))

        task = AsyncTask(
            self, fut, cancel_timeout=cancel_timeout,
            bubble=bubble, master=master, name=name)
        self.logger.info('adding task `%s` to nursery `%s`', task, self)
        self._pending_tasks.append(task)
        return task



import asyncio
import time

import pytest

from mock import Mock

from traio import Nursery, TaskException


@pytest.mark.asyncio
async def test_empty():
    """Empty Nursery"""
    async with Nursery():
        pass


@pytest.mark.asyncio
async def test_simple():
    """Simple nursery with one execution"""
    async def run(m):
        await asyncio.sleep(0.01)
        m()

    mock = Mock()

    async with Nursery() as n:
        n.start_soon(run(mock))

    assert mock.called


@pytest.mark.asyncio
async def test_block_raises():
    """Raise an exception from the block"""
    with pytest.raises(ValueError):
        async with Nursery():
            raise ValueError('boom')


@pytest.mark.asyncio
async def test_task_raises():
    """Raise an exception from a task"""
    async def raiser():
        await asyncio.sleep(0.01)
        raise ValueError('boom')

    try:
        async with Nursery() as n:
            n.start_soon(raiser())
    except TaskException as e:
        assert isinstance(e.__cause__, ValueError)
    else:
        raise Exception('DID NOT RAISE')


@pytest.mark.asyncio
async def test_join_task():
    """Test joining a task"""
    async def trivial():
        await asyncio.sleep(0.2)

    before = time.time()

    async with Nursery(timeout=0.5) as n:
        t = n.start_soon(trivial(), bubble=False)
        await t.join()
        # Cancel directly
        n.cancel()

    after = time.time()

    assert 0.1 < (after - before) < 0.5



"""
Test the `master` feature of `Nursery.start_soon`
"""

import asyncio
import time

import pytest

from tests import run10
from traio import Nursery, TaskException


@pytest.mark.asyncio
async def test_master():
    """Test if a master correctly cancels pending tasks"""
    async def master():
        await asyncio.sleep(0.01)

    before = time.time()

    async with Nursery() as n:
        n.start_soon(run10())
        n.start_soon(master(), master=True)

    after = time.time()

    assert (after - before) < 0.5


@pytest.mark.asyncio
async def test_master_raises():
    """Raise an exception from a task"""
    async def raiser():
        await asyncio.sleep(0.01)
        raise ValueError('boom')

    try:
        async with Nursery() as n:
            n.start_soon(run10())
            n.start_soon(raiser(), master=True)
    except TaskException as e:
        assert isinstance(e.__cause__, ValueError)
    else:
        raise Exception('DID NOT RAISE')


@pytest.mark.asyncio
async def test_cancel_master():
    """Test if cancelling master cancels the loop"""

    before = time.time()

    async with Nursery(timeout=0.5) as n:
        n.start_soon(run10())

        t = n.start_soon(run10(), master=True)

        await asyncio.sleep(0.05)
        t.cancel()

    after = time.time()

    assert (after - before) < 0.2
